# -*- coding=utf-8 -*-

"""Output of scaffolding subcommand."""

from pathlib import Path
from typing import Iterator, Literal, Optional

from revsymg.index_lib import IndexT, IndOrT, OrT

from khloraascaf.inputs import (
    FORWARD_STR,
    REVERSE_STR,
    STR_ORIENT,
    IdCT,
    OrStrT,
)
from khloraascaf.multiplied_doubled_contigs_graph import (
    CIND_IND,
    COR_IND,
    MDCGraphIDContainer,
)
from khloraascaf.result import ScaffoldingResult


# DOC: missing docstrings for constants
# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
OrCT = tuple[IdCT, OrT]

# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
#
# For OrCT type
#
ORC_ID_IND: Literal[0] = 0
ORC_OR_IND: Literal[1] = 1

# ---------------------------------------------------------------------------- #
#                                 Orientations                                 #
# ---------------------------------------------------------------------------- #
#
# Transformed orientation integer -> string
#
ORIENT_INT_STR: tuple[OrStrT, OrStrT] = (FORWARD_STR, REVERSE_STR)

# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Contigs of the Regions
#
CONTIGS_OF_REGIONS_PREFIX: Literal['contigs_of_regions'] = 'contigs_of_regions'
CONTIGS_OF_REGIONS_EXT: Literal['tsv'] = 'tsv'
#
# Map of the regions
#
MAP_OF_REGIONS_PREFIX: Literal['map_of_regions'] = 'map_of_regions'
MAP_OF_REGIONS_EXT: Literal['tsv'] = 'tsv'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Write                                    #
# ---------------------------------------------------------------------------- #
def write_contigs_of_regions(
        result: ScaffoldingResult, contigs_of_regions_path: Path,
        id_container: Optional[MDCGraphIDContainer] = None):
    """Write the regions' contigs file.

    # DOC: describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Previous scaffolding result
    contigs_of_regions_path : Path
        Path of the future file containing the regions' oriented contigs
    id_container : MDCGraphIDContainer, optional
        Identifiers container for the graph, by default None
    """
    with open(contigs_of_regions_path, 'w', encoding='utf-8') as f_out:
        for region_index in range(result.number_regions()):

            line: str = f'{region_index}'
            for occorc in result.region_occorc(region_index):
                if id_container is not None:
                    line += '\t' + str(
                        id_container.vertex_to_contig(occorc[CIND_IND]),
                    )
                else:
                    line += f'\t{occorc[CIND_IND]}'
                line += f'\t{ORIENT_INT_STR[occorc[COR_IND]]}'
            f_out.write(line + '\n')


def write_map_of_regions(result: ScaffoldingResult,
                         map_of_regions_path: Path):
    """Write the region map file.

    # DOC: describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Scaffolding result
    map_of_regions_path : Path
        Path of the future file containing the regions' oriented contigs
    """
    with open(map_of_regions_path, 'w', encoding='utf-8') as f_out:
        for reg_ind, reg_or in result.map_of_regions():
            f_out.write(f'{reg_ind}\t{ORIENT_INT_STR[reg_or]}\n')


# ---------------------------------------------------------------------------- #
#                                     Read                                     #
# ---------------------------------------------------------------------------- #
def read_contigs_of_regions(contigs_of_regions_path: Path) -> (
        Iterator[tuple[IndexT, list[OrCT]]]):
    """Read the regions' contigs file.

    Parameters
    ----------
    contigs_of_regions_path : Path
        List of oriented contigs for each region

    Yields
    ------
    IndexT
        Index of the region
    list of OrCT
        List of oriented contigs of the region
    """
    with open(contigs_of_regions_path, 'r', encoding='utf-8') as cor_in:
        for region_ind, line in enumerate(cor_in):
            orc_of_region: list[OrCT] = []
            l_orc = line.split()[1:]
            k = 0
            while k < len(l_orc) - 1:
                orc_of_region.append(
                    (l_orc[k], STR_ORIENT[l_orc[k + 1]]),  # type: ignore
                )
                k += 2
            yield region_ind, orc_of_region


def read_map_of_regions(map_of_regions_path: Path) -> Iterator[IndOrT]:
    """Write the region map file.

    Parameters
    ----------
    map_of_regions_path : Path
        Map of regions

    Yields
    ------
    IndOrT
        The index and the orientation of the region
    """
    with open(map_of_regions_path, 'r', encoding='utf-8') as mof_in:
        for line in mof_in:
            reg_indstr, reg_orstr = line.split()
            yield IndexT(reg_indstr), STR_ORIENT[reg_orstr]  # type: ignore


# ---------------------------------------------------------------------------- #
#                                Format Filename                               #
# ---------------------------------------------------------------------------- #
def fmt_contigs_of_regions_filename(instance_name: str) -> str:
    """Format the contigs of regions filename.

    Parameters
    ----------
    instance_name : str
        Instance name

    Returns
    -------
    str
        Formatted filename
    """
    return (f'{CONTIGS_OF_REGIONS_PREFIX}_{instance_name}'
            f'.{CONTIGS_OF_REGIONS_EXT}')


def fmt_map_of_regions_filename(instance_name: str) -> str:
    """Format map of the regions filename.

    Parameters
    ----------
    instance_name : str
        Instance name

    Returns
    -------
    str
        Formatted filename
    """
    return f'{MAP_OF_REGIONS_PREFIX}_{instance_name}.{MAP_OF_REGIONS_EXT}'



# -*- coding=utf-8 -*-

"""Utilitary module for debugging."""


from pathlib import Path
from typing import Iterator

from revsymg.index_lib import IndexT

from khloraascaf.multiplied_doubled_contigs_graph import OccOrCT
from khloraascaf.result import ScaffoldingResult


# DOC: module debug
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Vertices of the regions
#
VERTICES_OF_REGIONS_PREFIX = 'vertices_of_regions'
VERTICES_OF_REGIONS_EXT = 'tsv'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Write                                    #
# ---------------------------------------------------------------------------- #
def write_vertices_of_regions(result: ScaffoldingResult,
                              vertices_of_regions_path: Path):
    """Write the regions' vertex file.

    # DOC: describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Previous scaffolding result
    vertices_of_regions_path : Path
        Path of the future file containing the regions' vertices
    """
    with open(vertices_of_regions_path, 'w', encoding='utf-8') as f_out:
        for region_index in range(result.number_regions()):

            line: str = f'{region_index}'
            for occorc in result.region_occorc(region_index):
                line += f'\t{occorc}'
            f_out.write(line + '\n')


# ---------------------------------------------------------------------------- #
#                                     Read                                     #
# ---------------------------------------------------------------------------- #
def read_vertices_of_regions(vertices_of_regions_path: Path) -> (
        Iterator[tuple[IndexT, list[OccOrCT]]]):
    """Read the regions' vertices file.

    Parameters
    ----------
    vertices_of_regions_path : Path
        Path of the future file containing the regions' vertices

    Yields
    ------
    IndexT
        Index of the region
    list of OccOrCT
        List of vertices of the region
    """
    # TOTEST: read_vertices_of_regions
    with open(vertices_of_regions_path, 'r', encoding='utf-8') as cor_in:
        for region_ind, line in enumerate(cor_in):
            l_occorcstr = line.split('\t')[1:]
            vertices_of_region: list[OccOrCT] = [
                eval(occorcstr)
                for occorcstr in l_occorcstr
            ]
            yield region_ind, vertices_of_region


# ---------------------------------------------------------------------------- #
#                                Format Filename                               #
# ---------------------------------------------------------------------------- #
def fmt_vertices_of_regions_filename(instance_name: str) -> str:
    """Format the contigs of regions filename.

    Parameters
    ----------
    instance_name : str
        Instance name

    Returns
    -------
    str
        Formatted filename
    """
    return (f'{VERTICES_OF_REGIONS_PREFIX}_{instance_name}'
            f'.{VERTICES_OF_REGIONS_EXT}')



# -*- coding=utf-8 -*-

"""Module for assembly graph and multiple solution generator."""


from pathlib import Path
from queue import LifoQueue
from typing import Iterator

from bitarray import bitarray
from revsymg.graphs import RevSymGraph
from revsymg.index_lib import (
    FORWARD_INT,
    IND,
    ORIENT_REV,
    EIndOrIndT,
    IndexT,
    IndOrT,
    OrT,
)

from khloraascaf.outputs import (
    ORC_ID_IND,
    ORC_OR_IND,
    OrCT,
    read_contigs_of_regions,
    read_map_of_regions,
)


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
class AssemblyGraph():
    """Khloraascaf assembly graph."""

    def __init__(self, map_of_regions_path: Path,
                 contigs_of_regions_path: Path):
        """The Initializer."""
        self.__graph: RevSymGraph = RevSymGraph()
        self.__regions_contigs: list[list[OrCT]] = []
        self.__read_regions_contigs(contigs_of_regions_path)
        self.__add_regions_links(map_of_regions_path)

    def all_region_paths(self) -> Iterator[list[IndOrT]]:
        """Iterate over all the paths of oriented regions.

        Yields
        ------
        list of IndOrT
            Path of oriented regions
        """
        edges = self.__graph.edges()

        used_edges: bitarray = bitarray('0') * (
            edges.biggest_edge_index() + 1)

        starter = (0, FORWARD_INT)

        path: list[IndOrT] = [starter]
        eind_path: list[IndexT] = []
        lifo: LifoQueue[EIndOrIndT] = LifoQueue()

        for v, e_ind in edges.succs(starter):
            lifo.put((starter, v, e_ind))

        while not lifo.empty():
            _, v, e_ind = lifo.get()
            path.append(v)
            eind_path.append(e_ind)
            used_edges[e_ind] = True

            vw_not_used: list[EIndOrIndT] = []
            for w, f_ind in edges.succs(v):
                if w[IND] != starter[IND] and not used_edges[f_ind]:
                    vw_not_used.append((v, w, f_ind))

            if vw_not_used:
                for v, w, f_ind in vw_not_used:
                    lifo.put((v, w, f_ind))
            else:
                yield path
                #
                # Back to the first branching vertex
                #
                if not lifo.empty():
                    branch_src, branch_succ, branch_eind = lifo.get()
                    lifo.put((branch_src, branch_succ, branch_eind))
                    while path[-1] != branch_src:
                        path.pop()
                        used_edges[eind_path.pop()] = False

    def all_oriented_contig_paths(self) -> Iterator[list[OrCT]]:
        """Iterate over all the paths of oriented contigs.

        Yields
        ------
        list of IndOrT
            Path of oriented contigs
        """
        for region_path in self.all_region_paths():
            oriented_contig_path: list[OrCT] = []
            for reg_ind, reg_or in region_path:
                for orc in self.oriented_contigs_of_region(reg_ind, reg_or):
                    oriented_contig_path.append(orc)
            yield oriented_contig_path

    # ~*~ Getter ~*~

    def revsymg(self) -> RevSymGraph:
        """Return the reverse symmetric graph associated.

        Returns
        -------
        RevSymGraph
            Reverse symmetric graph
        """
        return self.__graph

    def oriented_contigs_of_region(self, region_ind: IndexT,
                                   orientation: OrT = FORWARD_INT) -> (
            Iterator[OrCT]):
        """Iterate over oriented contigs of the oriented region.

        Parameters
        ----------
        region_ind : IndexT
            Region's index
        orientation : OrT, optional
            Region's orientation, by default FORWARD_INT

        Yields
        ------
        OrCT
            Oriented contig of the oriented region
        """
        if orientation == FORWARD_INT:
            yield from self.__regions_contigs[region_ind]
        else:
            for oriented_contig in reversed(self.__regions_contigs[region_ind]):
                yield rev_oriented_contig(oriented_contig)

    # ~*~ Private ~*~

    def __read_regions_contigs(self, contigs_of_regions_path: Path):
        """Extract the oriented contigs for each region.

        Parameters
        ----------
        contigs_of_regions_path : Path
            List of oriented contigs for each region
        """
        vertices = self.__graph.vertices()
        for region_ind, oriented_contigs in read_contigs_of_regions(
                contigs_of_regions_path):
            assert region_ind == vertices.add()
            self.__regions_contigs.append(oriented_contigs)

    def __add_regions_links(self, map_of_regions_path: Path):
        """Add regions links from map of regions.

        Parameters
        ----------
        map_of_regions_path : Path
            Map of regions
        """
        edges = self.__graph.edges()
        reg_indor_it = read_map_of_regions(map_of_regions_path)
        start_indor = next(reg_indor_it)
        u_indor = start_indor
        for v_indor in reg_indor_it:
            edges.add(u_indor, v_indor)
            u_indor = v_indor
        edges.add(u_indor, start_indor)


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def rev_oriented_contig(oriented_contig: OrCT) -> OrCT:
    """Return the reverse of the oriented contig.

    Parameters
    ----------
    oriented_contig : OrCT
        Oriented contig

    Returns
    -------
    OrCT
        Its reverse
    """
    return oriented_contig[ORC_ID_IND], ORIENT_REV[oriented_contig[ORC_OR_IND]]



# -*- coding=utf-8 -*-

"""Scaffolding module."""

from copy import deepcopy
from pathlib import Path
from typing import Optional

from pulp import LpProblem, LpStatusOptimal

from khloraascaf.exceptions import (
    CannotPrioritiseILP,
    UnfeasibleDR,
    UnfeasibleIR,
    UnfeasibleUNIQUE,
)
from khloraascaf.ilp import solve_pulp_problem
from khloraascaf.ilp.pulp_max_dirf import intersected_dirf_model
from khloraascaf.ilp.pulp_max_invf import nested_invf_model
from khloraascaf.ilp.pulp_max_presscore import best_presscore_model
from khloraascaf.ilp.pulp_var_db import PuLPVarModelT
from khloraascaf.inputs import (
    INSTANCE_NAME_DEF,
    MULT_UPB_DEF,
    OUTDEBUG_DEF,
    OUTDIR_DEF,
    PRESSCORE_UPB_DEF,
    SOLVER_CBC,
    IdCT,
    MultT,
    PresScoreT,
)
from khloraascaf.multiplied_doubled_contigs_graph import (
    MDCGraph,
    OccOrCT,
    first_forward_occurence,
    mdcg_with_id_from_input_files,
)
from khloraascaf.outputs import (
    fmt_contigs_of_regions_filename,
    fmt_map_of_regions_filename,
    write_contigs_of_regions,
    write_map_of_regions,
)
from khloraascaf.result import ScaffoldingResult, path_to_regions
from khloraascaf.utils_debug import (
    fmt_vertices_of_regions_filename,
    write_vertices_of_regions,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
#
# Solutions codes
#
IR_CODE_SUFFIX = 'ir'
DR_CODE_SUFFIX = 'dr'
UNIQUE_CODE_SUFFIX = 'un'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                From Input File                               #
# ---------------------------------------------------------------------------- #
# pylint: disable=too-many-arguments
def scaffolding(
        contigs_attrs: Path,
        contigs_links: Path,
        contig_starter: IdCT,
        multiplicity_upperbound: MultT = MULT_UPB_DEF,
        presence_score_upperbound: PresScoreT = PRESSCORE_UPB_DEF,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF):
    """Computes the scaffolding.

    Parameters
    ----------
    contigs_attrs : Path
        Contigs' attributes file path
    contigs_links : Path
        Contigs' links file path
    contig_starter : IdCT
        Starter contig's identifier
    multiplicity_upperbound : MultT, optional
        Multiplicities upper bound, by default MULT_UPB_DEF
    presence_score_upperbound : PresScoreT, optional
        Multiplicities upper bound, by default MULT_UPB_DEF
    solver : str, optional
        MILP solver to use (CBC or GUROBI), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance name, by default INSTANCE_NAME_DEF
    debug : bool, optional
        Output debug or not, by default False
    """
    # DOC: verify docstring
    mdcg, id_container = mdcg_with_id_from_input_files(
        contigs_attrs, contigs_links,
        multiplicity_upperbound=multiplicity_upperbound,
        presence_score_upperbound=presence_score_upperbound,
    )
    starter_vertex = first_forward_occurence(
        id_container.contig_to_vertex(contig_starter))

    # ------------------------------------------------------------------------ #
    # Find the solutions
    # ------------------------------------------------------------------------ #
    # TODO: what to do if several optimal solutions?
    #   * solution should be a list of results?
    #   * how to tell which results was selected?
    #   * metadata on the outputs (human reading)?
    solution = combine_scaffolding_problems(
        mdcg,
        starter_vertex,
        solver=solver,
        outdir=outdir,
        instance_name=instance_name,
        debug=debug,
    )

    # ------------------------------------------------------------------------ #
    # Write the solution in files
    # ------------------------------------------------------------------------ #
    #
    # The order of oriented contigs for each region
    #
    write_contigs_of_regions(
        solution,
        id_container=id_container,
        contigs_of_regions_path=(
            outdir / fmt_contigs_of_regions_filename(instance_name)
        ),
    )
    #
    # The order of oriented regions
    #
    write_map_of_regions(
        solution,
        map_of_regions_path=(
            outdir / fmt_map_of_regions_filename(instance_name)
        ),
    )


# ---------------------------------------------------------------------------- #
#                            From Mathematical Data                            #
# ---------------------------------------------------------------------------- #
def combine_scaffolding_problems(
        mdcg: MDCGraph,
        starter_vertex: OccOrCT,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF) -> ScaffoldingResult:
    """Find a priority between the optimisation problems.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    solver : str, optional
        MILP solver to use (CBC or GUROBI), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance's name, by default INSTANCE_NAME_DEF
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    ScaffoldingResult
        Best ILP result

    Raises
    ------
    CannotPrioritiseILP
        Choosing the best solution is not possible
    """
    ir_solution = scaffolding_inverted_repeats(
        mdcg,
        starter_vertex,
        solver=solver,
        outdir=outdir,
        instance_name=f'{instance_name}_{IR_CODE_SUFFIX}',
        debug=debug,
    )

    dr_solution = scaffolding_directed_repeats(
        mdcg,
        starter_vertex,
        solver=solver,
        outdir=outdir,
        instance_name=f'{instance_name}_{DR_CODE_SUFFIX}',
        debug=debug,
    )

    # ------------------------------------------------------------------------ #
    # First level of scaffolding
    # ------------------------------------------------------------------------ #
    if ir_solution.opt_value() > dr_solution.opt_value():
        return scaffolding_unique(
            mdcg,
            starter_vertex,
            result=ir_solution,
            solver=solver,
            outdir=outdir,
            instance_name=(
                f'{instance_name}'
                f'_{IR_CODE_SUFFIX}_{UNIQUE_CODE_SUFFIX}'
            ),
            debug=debug,
        )

    if ir_solution.opt_value() < dr_solution.opt_value():
        return scaffolding_unique(
            mdcg,
            starter_vertex,
            result=dr_solution,
            solver=solver,
            outdir=outdir,
            instance_name=(
                f'{instance_name}'
                f'_{DR_CODE_SUFFIX}_{UNIQUE_CODE_SUFFIX}'
            ),
            debug=debug,
        )

    if ir_solution.opt_value() == 0:  # pylint: disable=compare-to-zero
        return scaffolding_unique(
            mdcg,
            starter_vertex,
            solver=solver,
            outdir=outdir,
            instance_name=(
                f'{instance_name}'
                f'_{UNIQUE_CODE_SUFFIX}'
            ),
            debug=debug,
        )

    # TODO: should repeat to construct solution until no changes?
    #   * ir
    #       * ir - ir
    #           * ir - ir - ir
    #           * ir - ir - dr
    #               * ir - ir - unique
    #       * ir - dr
    #   * dr
    #       * dr - ir
    #       * dr - dr
    #
    # ------------------------------------------------------------------------ #
    # Second level of scaffolding
    # ------------------------------------------------------------------------ #
    ir_unique_solution = scaffolding_unique(
        mdcg,
        starter_vertex,
        result=ir_solution,
        solver=solver,
        outdir=outdir,
        instance_name=(
            f'{instance_name}'
            f'_{IR_CODE_SUFFIX}_{UNIQUE_CODE_SUFFIX}'
        ),
        debug=debug,
    )

    dr_unique_solution = scaffolding_unique(
        mdcg,
        starter_vertex,
        result=dr_solution,
        solver=solver,
        outdir=outdir,
        instance_name=(
            f'{instance_name}'
            f'_{DR_CODE_SUFFIX}_{UNIQUE_CODE_SUFFIX}'
        ),
        debug=debug,
    )

    if ir_unique_solution.opt_value() > dr_unique_solution.opt_value():
        return ir_unique_solution
    if ir_unique_solution.opt_value() < dr_unique_solution.opt_value():
        return dr_unique_solution

    # TODO: third level of scaffolding (ir-dr-unique) & (dr-ir-unique)

    # FIXME: show different auto with instance_name, or manually?
    #   * fill a list of tuple (name, value)
    raise CannotPrioritiseILP(
        [
            (res.instance_name(), res.opt_value())
            for res in (ir_solution,
                        dr_solution,
                        ir_unique_solution,
                        dr_unique_solution)
        ],
    )


# ---------------------------------------------------------------------------- #
#                         Specific Regions Scaffolding                         #
# ---------------------------------------------------------------------------- #
def scaffolding_inverted_repeats(
        mdcg: MDCGraph,
        starter_vertex: OccOrCT,
        result: Optional[ScaffoldingResult] = None,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF) -> ScaffoldingResult:
    """Find the best inverted repeats.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    result : ScaffoldingResult, optional
        Previous scaffolding result
    solver : str, optional
        MILP solver to use (CBC or GUROBI), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance's name, by default INSTANCE_NAME_DEF
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    ScaffoldingResult
        Scaffolding result

    Raises
    ------
    UnfeasibleIR
        The combinatorial problem is unfeasible
    """
    prob, var = nested_invf_model(mdcg, starter_vertex, result=result)
    solve_pulp_problem(
        prob,
        solver=solver,
        outdir=outdir,
        instance_name=instance_name,
    )
    if prob.status != LpStatusOptimal:
        raise UnfeasibleIR()

    return __fill_the_result(
        prob, var, mdcg, starter_vertex,
        outdir, instance_name, debug,
        result=result,
    )


def scaffolding_directed_repeats(
        mdcg: MDCGraph,
        starter_vertex: OccOrCT,
        result: Optional[ScaffoldingResult] = None,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF) -> ScaffoldingResult:
    """Find the best directed repeats.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    result : ScaffoldingResult, optional
        Previous scaffolding result
    solver : str, optional
        MILP solver to use (CBC or GUROBI), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance's name, by default INSTANCE_NAME_DEF
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    ScaffoldingResult
        Scaffolding result

    Raises
    ------
    UnfeasibleDR
        The combinatorial problem is unfeasible
    """
    prob, var = intersected_dirf_model(
        mdcg, starter_vertex,
        result=result,
    )
    solve_pulp_problem(
        prob,
        solver=solver,
        outdir=outdir,
        instance_name=instance_name,
    )
    if prob.status != LpStatusOptimal:
        raise UnfeasibleDR()

    return __fill_the_result(
        prob, var, mdcg, starter_vertex,
        outdir, instance_name, debug,
        result=result,
    )


def scaffolding_unique(
        mdcg: MDCGraph,
        starter_vertex: OccOrCT,
        result: Optional[ScaffoldingResult] = None,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF) -> ScaffoldingResult:
    """Find the best unique region.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    result : ScaffoldingResult
        Previous scaffolding result
    solver : str, optional
        MILP solver to use (CBC or GUROBI), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance's name, by default INSTANCE_NAME_DEF
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    ScaffoldingResult
        Scaffolding result

    Raises
    ------
    UnfeasibleUNIQUE
        The combinatorial problem is unfeasible
    """
    prob, var = best_presscore_model(mdcg, starter_vertex, result=result)
    solve_pulp_problem(
        prob,
        solver=solver,
        outdir=outdir,
        instance_name=instance_name,
    )
    if prob.status != LpStatusOptimal:
        raise UnfeasibleUNIQUE()

    return __fill_the_result(
        prob, var, mdcg, starter_vertex,
        outdir, instance_name, debug,
        result=result,
    )


def __fill_the_result(
        prob: LpProblem,
        var: PuLPVarModelT,
        mdcg: MDCGraph,
        starter_vertex: OccOrCT,
        outdir: Path,
        instance_name: str,
        debug: bool,
        result: Optional[ScaffoldingResult] = None) -> ScaffoldingResult:
    """Find the best unique region.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    outdir : Path
        Output directory path
    instance_name : str
        Instance's name
    debug : bool
        Output debug or not
    result : ScaffoldingResult, optional
        Previous scaffolding result, by default None

    Returns
    -------
    ScaffoldingResult
        Scaffolding result

    Raises
    ------
    UnfeasibleUNIQUE
        The combinatorial problem is unfeasible
    """
    if result is None:
        new_result = ScaffoldingResult(instance_name)
    else:
        new_result = deepcopy(result)
        new_result.set_instance_name(instance_name)
    if prob.objective.value() is not None:
        # see PuLP issue 331
        new_result.set_opt_value(prob.objective.value())

    # TODO: think about filling result with already existing result
    path_to_regions(mdcg, starter_vertex, var, new_result)

    # ------------------------------------------------------------------------ #
    # Debug
    # ------------------------------------------------------------------------ #
    if debug:
        #
        # The order of vertices for each region
        #
        write_vertices_of_regions(
            new_result,
            vertices_of_regions_path=(
                outdir
                / fmt_vertices_of_regions_filename(instance_name)
            ),
        )
        write_map_of_regions(
            new_result,
            map_of_regions_path=(
                outdir
                / fmt_map_of_regions_filename(instance_name)
            ),
        )
    return new_result



# -*- coding=utf-8 -*-

"""Solve integer lieanr programs module."""


from __future__ import division

from pathlib import Path

from pulp import GUROBI_CMD, PULP_CBC_CMD, LpProblem

from khloraascaf.exceptions import WrongSolverName
from khloraascaf.inputs import (
    INSTANCE_NAME_DEF,
    OUTDIR_DEF,
    SOLVER_CBC,
    SOLVER_GUROBI,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                      CBC                                     #
# ---------------------------------------------------------------------------- #
LOG_CBC_PREFIX = 'cbc'
LOG_CBC_EXT = 'log'

# ---------------------------------------------------------------------------- #
#                                    Gurobi                                    #
# ---------------------------------------------------------------------------- #
LOG_GUROBI_PREFIX = 'gurobi'
LOG_GUROBI_EXT = 'log'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                 Solve Models                                 #
# ---------------------------------------------------------------------------- #
def solve_pulp_problem(prob: LpProblem,
                       solver: str = SOLVER_CBC,
                       outdir: Path = OUTDIR_DEF,
                       instance_name: str = INSTANCE_NAME_DEF):
    """Instanciate and solve the PuLP model.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    solver : str, optional
        MILP solver to use ('CBC' or 'GUROBI'), by default SOLVER_CBC
    outdir : Path, optional
        Output directory path, by default OUTDIR_DEF
    instance_name : str, optional
        Instance's name, by default INSTANCE_NAME_DEF

    Raises
    ------
    WrongSolverName
        The solver name is not correct
    """
    # REFACTOR: remove default arguments?
    if solver == SOLVER_CBC:
        prob.solve(
            PULP_CBC_CMD(
                msg=0,
                logPath=(outdir / fmt_cbc_log_name(instance_name)),
            ),
        )
    elif solver == SOLVER_GUROBI:
        prob.solve(
            GUROBI_CMD(
                msg=0,
                logPath=(outdir / fmt_gurobi_log_name(instance_name)),
            ),
        )
    else:
        raise WrongSolverName(solver)


# ---------------------------------------------------------------------------- #
#                                Logs Formatters                               #
# ---------------------------------------------------------------------------- #
def fmt_gurobi_log_name(instance_name: str) -> str:
    """Format Gurobi log file path.

    Parameters
    ----------
    instance_name : str
        Instance name

    Returns
    -------
    str
        Formatted filename
    """
    return f'{instance_name}_{LOG_GUROBI_PREFIX}.{LOG_GUROBI_EXT}'


def fmt_cbc_log_name(instance_name: str) -> str:
    """Format CBC log file path.

    Parameters
    ----------
    instance_name : str
        Instance name

    Returns
    -------
    str
        Formatted filename
    """
    return f'{instance_name}_{LOG_CBC_PREFIX}.{LOG_CBC_EXT}'



# -*- coding=utf-8 -*-

"""Caffolding result's module: structures and utils."""


from __future__ import annotations

import sys
from queue import LifoQueue, Queue
from typing import Iterator, Optional

from revsymg.index_lib import FORWARD_INT, REVERSE_INT, IndexT, OrT

from khloraascaf.exceptions import NotACircuit
from khloraascaf.ilp.dirf_sets import dirf_canonical, dirf_other
from khloraascaf.ilp.invf_sets import invf_canonical, invf_other
from khloraascaf.ilp.pulp_var_db import (
    BIN_THRESHOLD,
    PuLPVarDirFModel,
    PuLPVarInvFModel,
    PuLPVarModelT,
)
from khloraascaf.multiplied_doubled_contigs_graph import (
    COCC_IND,
    COR_IND,
    MDCGraph,
    OccOrCT,
)


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                Genomic Regions                               #
# ---------------------------------------------------------------------------- #
GenRegionMapT = list[tuple[IndexT, OrT]]
GenRegionT = list[OccOrCT]


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
class ScaffoldingResult():
    """Scaffolding result class.

    An instance contains a genomic regions map and each genomic regions'
    oriented contigs.

    Warning
    -------
    The result corresponds to the linear path in mdcg so the circularity is not
    fixed here, i.e. the first region begins by mdcg.occorc_start() and the last
    region finishes by mdcg.occorc_terminal()
    """

    # DOC: add to docstring example to get regions' orc from result

    __NO_OPT_VALUE = 0

    def __init__(self, instance_name: str):
        """The Initializer."""
        self.__instance_name = instance_name
        #
        # Genomic region
        #
        self.__regions_map: GenRegionMapT = []
        self.__regions_occorc: list[GenRegionT] = []
        self.__unique_regions: list[IndexT] = []
        self.__ir_regions: list[IndexT] = []
        self.__dr_regions: list[IndexT] = []
        #
        # ILP optimal value
        #
        self.__opt_value: int = self.__NO_OPT_VALUE

    # ~*~ Setter ~*~

    def set_opt_value(self, opt_value: int):
        """Set ILP optimal value.

        Parameters
        ----------
        opt_value : int
            The optimal value to set
        """
        self.__opt_value = opt_value

    def clear_regions(self):
        """Delete the current regions result."""
        self.__regions_map.clear()
        self.__regions_occorc.clear()
        self.__unique_regions.clear()
        self.__ir_regions.clear()
        self.__dr_regions.clear()

    def add_region(self, region_code: int,
                   region_index: Optional[IndexT] = None) -> IndexT:
        """Add a region.

        Parameters
        ----------
        region_code : int
            Region code (0: UNIQUE; 1: IR; 2: DR)
        region_index : IndexT, optional
            Index of an already existing region, else None

        Returns
        -------
        IndexT
            Region's index
        """
        #
        # Non-existing region
        #
        if region_index is None:
            region_index = len(self.__regions_occorc)
            self.__regions_occorc.append([])
            self.__regions_map.append((region_index, FORWARD_INT))
            if region_code == 0:  # pylint: disable=compare-to-zero
                self.__unique_regions.append(region_index)
            elif region_code == 1:
                self.__ir_regions.append(region_index)
            elif region_code == 2:
                self.__dr_regions.append(region_index)
            else:
                # TODO: exception wrong region code
                sys.exit('ERROR: wrong region code')
        #
        # Already existing region (repeat)
        #
        else:
            if region_code == 0:  # pylint: disable=compare-to-zero
                # TODO: exception can't add twice a UNIQ region
                sys.exit("ERROR: can't add twice a UNIQ region")
            if region_code == 1:
                self.__regions_map.append((region_index, REVERSE_INT))
            elif region_code == 2:
                self.__regions_map.append((region_index, FORWARD_INT))
            else:
                # TODO: exception wrong region code
                sys.exit('ERROR: wrong region code')
        return region_index

    def add_occorc_to_region(self, v: OccOrCT, region_index: IndexT):
        """Add v to the region denoted by its index.

        Parameters
        ----------
        v : OccOrCT
            Multiplied oriented contig
        region_index : IndexT
            Region's index
        """
        self.__regions_occorc[region_index].append(v)

    def set_instance_name(self, instance_name: str):
        """Set instance name.

        Parameters
        ----------
        instance_name : str
            Instance name
        """
        self.__instance_name = instance_name

    # ~*~ Getter ~*~

    def instance_name(self) -> str:
        """Return the instance name.

        Returns
        -------
        str
            Instance name
        """
        return self.__instance_name

    def opt_value(self) -> int:
        """ILP optimal value.

        Returns
        -------
        int
            ILP optimal value
        """
        return self.__opt_value

    def number_regions(self) -> int:
        """Returns the number of regions.

        Returns
        -------
        int
            Number of regions
        """
        return len(self.__regions_occorc)

    def unique_regions(self) -> Iterator[IndexT]:
        """Iterate over unique regions index.

        Yields
        ------
        IndexT
            Unique region index
        """
        yield from self.__unique_regions

    def ir_regions(self) -> Iterator[IndexT]:
        """Iterate over inverted repeats index.

        Yields
        ------
        IndexT
            Inverted repeats index
        """
        yield from self.__ir_regions

    def dr_regions(self) -> Iterator[IndexT]:
        """Iterate over directed repeats index.

        Yields
        ------
        IndexT
            Directed repeats index
        """
        yield from self.__dr_regions

    def map_of_regions(self) -> Iterator[tuple[IndexT, OrT]]:
        """Iterate over the regions and their orientation.

        Yields
        ------
        IndexT
            Region's index
        OrT
            Region's orientation
        """
        yield from self.__regions_map

    def region_occorc(self, region_index: IndexT) -> Iterator[OccOrCT]:
        """Iterate over the multiplied oriented contig of the region.

        Parameters
        ----------
        region_index : IndexT
            Region's index

        Yields
        ------
        OccOrCT
            Multiplied oriented contig of the region
        """
        yield from self.__regions_occorc[region_index]


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                Fill The Result                               #
# ---------------------------------------------------------------------------- #
def path_to_regions(mdcg: MDCGraph,
                    starter_vertex: OccOrCT,
                    var: PuLPVarModelT,
                    result: ScaffoldingResult):
    """Extract regions from optimal path.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    var : PuLPVarModelT
        PuLP variables
    result : ScaffoldingResult
        Scaffolding result with regions
    """
    # ------------------------------------------------------------------------ #
    # Manage paired fragments
    # ------------------------------------------------------------------------ #
    # TODO: think about e.g. several IR & how to adapt these functions bellow?
    #   * if the pulp model want to keep subpath with avoiding the previous
    #       repeated fragments to be paired, then is it sufficient to read
    #       previous repeated region in result object?
    set_invf_paired = __create_set_invf_paired(var, result)
    set_dirf_paired = __create_set_dirf_paired(var, result)

    # ------------------------------------------------------------------------ #
    # Init regions
    # ------------------------------------------------------------------------ #
    result.clear_regions()
    # reg_code:
    #   - 0: UNIQUE
    #   - 1: IR
    #   - 2: DR
    reg_code = 0
    region_index = result.add_region(reg_code)

    initial_vertex = __find_initial(
        mdcg, starter_vertex, var, set_invf_paired, set_dirf_paired)

    u: OccOrCT = initial_vertex  # previous v
    v: Optional[OccOrCT] = initial_vertex

    #
    # Region IR: LIFO, DR: FIFO
    #
    # XXX: no order between pairs of repeats
    #   - for the moment, the pairs of contiguous repeat are not considered
    #       ordered (e.g. pairs of contiguous IR should be lifo, DR fifo)
    ir_lifo: dict[IndexT, LifoQueue[OccOrCT]] = {}
    dr_fifo: dict[IndexT, Queue[OccOrCT]] = {}
    #
    # Canonical to the region's index: repeat was discovered
    #
    ir_canonical_reg: dict[OccOrCT, IndexT] = {}
    dr_canonical_reg: dict[OccOrCT, IndexT] = {}

    # ------------------------------------------------------------------------ #
    # Walk into the solution path
    # ------------------------------------------------------------------------ #
    while v is not None:
        new_reg_code = __get_region_code(v, set_invf_paired, set_dirf_paired)
        #
        # Type is contiguous
        #
        if new_reg_code == reg_code:
            #
            # Previous UNIQ region
            #
            if new_reg_code == 0:  # pylint: disable=compare-to-zero
                result.add_occorc_to_region(v, region_index)
            #
            # Previous or new IR region
            #
            elif new_reg_code == 1:
                canonical_invf = invf_canonical(v)
                #
                # Contiguous or new forward IR region
                #
                if canonical_invf not in ir_canonical_reg:
                    #
                    # Contiguous forward IR region
                    #
                    if __is_repeat_contiguous(u, v, var, new_reg_code):
                        result.add_occorc_to_region(v, region_index)
                        ir_lifo[region_index].put(invf_other(v))
                        ir_canonical_reg[canonical_invf] = region_index
                    #
                    # New forward IR region
                    #
                    else:
                        region_index = result.add_region(new_reg_code)
                        result.add_occorc_to_region(v, region_index)
                        ir_lifo[region_index] = LifoQueue()
                        ir_lifo[region_index].put(invf_other(v))
                        ir_canonical_reg[canonical_invf] = region_index

                #
                # IR region already existing
                #
                else:
                    #
                    # New reverse IR region
                    #
                    if region_index != ir_canonical_reg[canonical_invf]:
                        region_index = ir_canonical_reg[canonical_invf]
                        result.add_region(new_reg_code, region_index)
                    assert v == ir_lifo[region_index].get()
            #
            # Previous or new DR region
            #
            elif new_reg_code == 2:
                #
                # Contiguous or new first DR region
                #
                canonical_dirf = dirf_canonical(v)
                if canonical_dirf not in dr_canonical_reg:
                    #
                    # Contiguous first DR region
                    #
                    # FIXME: region_index provides the first repeat occorc
                    if __is_repeat_contiguous(u, v, var, new_reg_code):
                        result.add_occorc_to_region(v, region_index)
                        dr_fifo[region_index].put(dirf_other(v))
                        dr_canonical_reg[canonical_dirf] = region_index
                    #
                    # New first DR region
                    #
                    else:
                        region_index = result.add_region(new_reg_code)
                        result.add_occorc_to_region(v, region_index)
                        dr_fifo[region_index] = Queue()
                        dr_fifo[region_index].put(dirf_other(v))
                        dr_canonical_reg[canonical_dirf] = region_index

                #
                # DR region already existing
                #
                else:
                    #
                    # New second DR region
                    #
                    if region_index != dr_canonical_reg[canonical_dirf]:
                        region_index = dr_canonical_reg[canonical_dirf]
                        result.add_region(new_reg_code, region_index)
                    assert v == dr_fifo[region_index].get()
        #
        # New type of region
        #
        else:
            #
            # New UNIQ region
            #
            if new_reg_code == 0:  # pylint: disable=compare-to-zero
                region_index = result.add_region(new_reg_code)
                result.add_occorc_to_region(v, region_index)
            #
            # New or already existing IR region
            #
            elif new_reg_code == 1:
                canonical_invf = invf_canonical(v)
                #
                # New forward IR
                #
                if canonical_invf not in ir_canonical_reg:
                    region_index = result.add_region(new_reg_code)
                    result.add_occorc_to_region(v, region_index)
                    ir_lifo[region_index] = LifoQueue()
                    ir_lifo[region_index].put(invf_other(v))
                    ir_canonical_reg[canonical_invf] = region_index
                #
                # New reverse IR
                #
                else:
                    region_index = ir_canonical_reg[canonical_invf]
                    result.add_region(new_reg_code, region_index)
                    assert v == ir_lifo[region_index].get()
            #
            # New or already existing DR region
            #
            elif new_reg_code == 2:
                canonical_dirf = dirf_canonical(v)
                #
                # New first DR
                #
                if canonical_dirf not in dr_canonical_reg:
                    region_index = result.add_region(new_reg_code)
                    result.add_occorc_to_region(v, region_index)
                    dr_fifo[region_index] = Queue()
                    dr_fifo[region_index].put(dirf_other(v))
                    dr_canonical_reg[canonical_dirf] = region_index
                #
                # New second DR
                #
                else:
                    region_index = dr_canonical_reg[canonical_dirf]
                    result.add_region(new_reg_code, region_index)
                    assert v == dr_fifo[region_index].get()

        reg_code = new_reg_code
        u = v
        v = __succ_in_path(v, mdcg, var, initial_vertex)


# ---------------------------------------------------------------------------- #
#                               Walk In The Path                               #
# ---------------------------------------------------------------------------- #
def __pred_in_path(v: OccOrCT, mdcg: MDCGraph,
                   var: PuLPVarModelT, initial_vertex: OccOrCT) -> (
        Optional[OccOrCT]):
    """Return the predecessor of vertex v in solution path.

    Stop if the predecessor is the initial vertex.

    Parameters
    ----------
    v : OccOrCT
        Vertex
    mdcg : MDCGraph
        Multiplied doubled contig graph
    var : PuLPVarModelT
        PuLP variables
    initial_vertex : OccOrCT
        Starter vertex

    Returns
    -------
    OccOrCT, optional
        The previous vertex in solution path, None if it is the
        initial vertex

    Raises
    ------
    NotACircuit
        If the path is not a circuit
    """
    for u in mdcg.multiplied_preds(v):
        if var.x[u, v].varValue > BIN_THRESHOLD:
            if u == initial_vertex:
                return None
            return u
    raise NotACircuit()


def __succ_in_path(v: OccOrCT, mdcg: MDCGraph,
                   var: PuLPVarModelT, initial_vertex: OccOrCT) -> (
        Optional[OccOrCT]):
    """Return the successor of vertex v in solution path.

    Stop if the successor is the initial vertex.

    Parameters
    ----------
    v : OccOrCT
        Vertex
    mdcg : MDCGraph
        Multiplied doubled contig graph
    var : PuLPVarModelT
        PuLP variables
    initial_vertex : OccOrCT
        Starter vertex

    Returns
    -------
    OccOrCT, optional
        The next vertex in solution path, None if it is the
        initial vertex

    Raises
    ------
    NotACircuit
        If the path is not a circuit
    """
    for w in mdcg.multiplied_succs(v):
        if var.x[v, w].varValue > BIN_THRESHOLD:
            if w == initial_vertex:
                return None
            return w
    raise NotACircuit()


# ---------------------------------------------------------------------------- #
#                                Initialisation                                #
# ---------------------------------------------------------------------------- #
def __create_set_invf_paired(var: PuLPVarModelT,
                             result: ScaffoldingResult) -> set[OccOrCT]:
    """Create set of canonical of paired inverted fragments.

    Parameters
    ----------
    var : PuLPVarModelT
        PuLP variables
    result : ScaffoldingResult
        Scaffolding result

    Returns
    -------
    set of OccOrCT
        Set of canonical of paired inverted fragments
    """
    set_invf_paired: set[OccOrCT] = set()
    #
    # Add old inverted fragments pairing
    #
    for region_index in result.ir_regions():
        for v in result.region_occorc(region_index):
            set_invf_paired.add(invf_canonical(v))
    #
    # Add new inverted fragments pairing
    #
    if isinstance(var, PuLPVarInvFModel):
        for p, var_p in var.m.items():
            if var_p.value() > BIN_THRESHOLD:
                set_invf_paired.add(p[FORWARD_INT])
    return set_invf_paired


def __create_set_dirf_paired(var: PuLPVarModelT,
                             result: ScaffoldingResult) -> set[OccOrCT]:
    """Create set of canonical of paired directed fragments.

    Parameters
    ----------
    var : PuLPVarModelT
        PuLP variables
    result : ScaffoldingResult
        Scaffolding result

    Returns
    -------
    set of OccOrCT
        Set of canonical of paired directed fragments
    """
    set_dirf_paired: set[OccOrCT] = set()
    #
    # Add old directed fragments pairing
    #
    for region_index in result.dr_regions():
        for v in result.region_occorc(region_index):
            set_dirf_paired.add(dirf_canonical(v))
    #
    # Add new directed fragments pairing
    #
    if isinstance(var, PuLPVarDirFModel):
        for p, var_p in var.m.items():
            if var_p.value() > BIN_THRESHOLD:
                set_dirf_paired.add(p[FORWARD_INT])
    return set_dirf_paired


# pylint: disable=too-many-arguments
def __find_initial(mdcg: MDCGraph, starter_vertex: OccOrCT,
                   var: PuLPVarModelT,
                   set_invf_paired: set[OccOrCT],
                   set_dirf_paired: set[OccOrCT]) -> OccOrCT:
    """The first vertex of the starter's region.

    In case of circular unique region, it returns starter_vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    var : PuLPVarModelT
        PuLP variables
    set_invf_paired : set of OccOrCT
        Set of canonical of paired inverted fragments
    set_dirf_paired : set of OccOrCT
        Set of canonical of paired directed fragments

    Returns
    -------
    OccOrCT
        The first vertex of the starter's region
    """
    v: OccOrCT = starter_vertex
    u: Optional[OccOrCT] = __pred_in_path(
        starter_vertex, mdcg, var, starter_vertex)
    # pylint: disable=compare-to-zero
    while (u is not None
           and __get_region_code(u, set_invf_paired, set_dirf_paired) == 0
           ):
        v = u
        u = __pred_in_path(v, mdcg, var, starter_vertex)
    if u is None:
        return starter_vertex
    return v


# ---------------------------------------------------------------------------- #
#                               Region Management                              #
# ---------------------------------------------------------------------------- #
def __get_region_code(v: OccOrCT, set_invf_paired: set[OccOrCT],
                      set_dirf_paired: set[OccOrCT]) -> int:
    """Get the code of the region for the multiplied oriented contig.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig
    set_invf_paired : set of OccOrCT
        Set of canonical of paired inverted fragments
    set_dirf_paired : set of OccOrCT
        Set of canonical of paired directed fragments

    Returns
    -------
    int
        Code of the region (0: UNIQUE; 1: IR; 2: DR)
    """
    if dirf_canonical(v) in set_dirf_paired:
        return 2
    if ((v[COCC_IND] - v[COR_IND]) % 2 == 0  # pylint: disable=compare-to-zero
            and invf_canonical(v) in set_invf_paired):
        return 1
    return 0


def __is_repeat_contiguous(u: OccOrCT, v: OccOrCT,
                           var: PuLPVarModelT,
                           region_code: int) -> bool:
    """Answer if the repeat given by its code is contiguous.

    Parameters
    ----------
    u : OccOrCT
        Multiplied oriented contig
    v : OccOrCT
        Multiplied oriented contig
    var : PuLPVarModelT
        PuLP variables
    region_code : int
        Region's code

    Returns
    -------
    bool
        True if repeat is contiguous, else False
    """
    #
    # IR:
    #   i (= u) -> k(= v): ok, so is there l -> j?
    #
    if region_code == 1:
        return var.x[invf_other(v), invf_other(u)].varValue > BIN_THRESHOLD
    #
    # DR:
    #   i (= u) -> k(= v): ok, so is there j -> l?
    #
    if region_code == 2:
        return var.x[dirf_other(u), dirf_other(v)].varValue > BIN_THRESHOLD
    # TODO: error out of code
    return False



# -*- coding=utf-8 -*-

"""Scaffolding exceptions module."""

from typing import Any


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
class WrongSolverName(Exception):
    """Wrong solver name exception."""

    def __init__(self, solver: str):
        """The Initializer."""
        super().__init__()
        self.__solver = solver

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'ERROR: the solver name {self.__solver} is not correct\n'


# ---------------------------------------------------------------------------- #
#                                Prioritise ILP                                #
# ---------------------------------------------------------------------------- #
class CannotPrioritiseILP(Exception):
    """ILP prioritising exception."""

    def __init__(self, l_name_opt: list[tuple[str, Any]]):
        """The Initializer."""
        super().__init__()
        self.__l_name_opt = l_name_opt

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        msg = 'ERROR: impossible to prioritise the optimisation problems.\n'
        for name, value in self.__l_name_opt:
            msg += f'\t* {name} optimal value: {value}\n'
        return msg


# ---------------------------------------------------------------------------- #
#                                Unfeasible ILP                                #
# ---------------------------------------------------------------------------- #
class _UnfeasibleILP(Exception):
    """ILP problem unfeasible exception."""

    _PROBLEM_ID = 'THE PROBLEM'

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            'ERROR: the following problem is unfeasible:\n'
            f'\t{self._PROBLEM_ID}\n'
        )


class UnfeasibleIR(_UnfeasibleILP):
    """IR optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best inverted repeats'


class UnfeasibleDR(_UnfeasibleILP):
    """DR optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best directed repeats'


class UnfeasibleUNIQUE(_UnfeasibleILP):
    """UNIQUE optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best unique regions'


# ---------------------------------------------------------------------------- #
#                                 Result Error                                 #
# ---------------------------------------------------------------------------- #
class NotACircuit(Exception):
    """Not a circuit exception."""

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return 'ERROR: the found path is not a circuit'



# -*- coding=utf-8 -*-

"""Functional testing with toy examples.

See tests/data/README.md
"""

# pylint: disable=compare-to-zero, missing-raises-doc
from pathlib import Path
from typing import Iterator

from revsymg.graphs import RevSymGraph
from revsymg.index_lib import FORWARD_INT, REVERSE_INT, IndexT, IndOrT

from khloraascaf.assembly_graph import AssemblyGraph, OrCT, rev_oriented_contig
from khloraascaf.inputs import STR_ORIENT


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
TEST_DIR: Path = Path(__file__).parent.absolute()

_TOY_DATADIR = TEST_DIR / 'data'

# ---------------------------------------------------------------------------- #
#                                   IR Alone                                   #
# ---------------------------------------------------------------------------- #
_IR_ALONE_DIR = _TOY_DATADIR / 'IR_alone'
_IR_ALONE_SOL_REGMAP = _IR_ALONE_DIR / 'map_of_regions_sol_0.tsv'
_IR_ALONE_SOL_REGCTG_F = _IR_ALONE_DIR / 'contigs_of_regions_sol_0.tsv'
_IR_ALONE_REG_PATHS = _IR_ALONE_DIR / 'region_paths.tsv'
_IR_ALONE_ORC_PATHS = _IR_ALONE_DIR / 'oriented_contig_paths.tsv'

# ---------------------------------------------------------------------------- #
#                                   DR Alone                                   #
# ---------------------------------------------------------------------------- #
_DR_ALONE_DIR = _TOY_DATADIR / 'DR_alone'
_DR_ALONE_SOL_REGMAP = _DR_ALONE_DIR / 'map_of_regions_sol_0.tsv'
_DR_ALONE_SOL_REGCTG = _DR_ALONE_DIR / 'contigs_of_regions_sol_0.tsv'
_DR_ALONE_REG_PATHS = _DR_ALONE_DIR / 'region_paths.tsv'
_DR_ALONE_ORC_PATHS = _DR_ALONE_DIR / 'oriented_contig_paths.tsv'


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               All_region_paths                               #
# ---------------------------------------------------------------------------- #
def test_all_region_paths_ir():
    """Test all_region_paths method for inverted repeat."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_ALONE_SOL_REGMAP,
        _IR_ALONE_SOL_REGCTG_F,
    )
    s_region_paths: set[tuple[IndOrT, ...]] = {
        tuple(region_path)
        for region_path in read_region_paths(_IR_ALONE_REG_PATHS)
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_region_paths()]
    assert len(asm_graph_paths) == len(s_region_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_region_paths


def test_all_region_paths_dr():
    """Test all_region_paths method for directed repeat."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _DR_ALONE_SOL_REGMAP,
        _DR_ALONE_SOL_REGCTG,
    )
    s_region_paths: set[tuple[IndOrT, ...]] = {
        tuple(region_path)
        for region_path in read_region_paths(_DR_ALONE_REG_PATHS)
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_region_paths()]
    assert len(asm_graph_paths) == len(s_region_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_region_paths


# ---------------------------------------------------------------------------- #
#                           All_oriented_contig_paths                          #
# ---------------------------------------------------------------------------- #
def test_all_oriented_contig_paths_ir():
    """Test all_oriented_contig_paths method for IR."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_ALONE_SOL_REGMAP,
        _IR_ALONE_SOL_REGCTG_F,
    )
    s_orc_paths: set[tuple[OrCT, ...]] = {
        tuple(orc_path)
        for orc_path in read_oriented_contig_paths(_IR_ALONE_ORC_PATHS)
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_oriented_contig_paths()]
    assert len(asm_graph_paths) == len(s_orc_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_orc_paths


def test_all_oriented_contig_paths_dr():
    """Test all_oriented_contig_paths method for DR."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _DR_ALONE_SOL_REGMAP,
        _DR_ALONE_SOL_REGCTG,
    )
    s_orc_paths: set[tuple[OrCT, ...]] = {
        tuple(orc_path)
        for orc_path in read_oriented_contig_paths(_DR_ALONE_ORC_PATHS)
    }
    asm_graph_paths = [tuple(l) for l in asm_graph.all_oriented_contig_paths()]
    assert len(asm_graph_paths) == len(s_orc_paths)
    for reg_path in asm_graph_paths:
        assert reg_path in s_orc_paths


# ---------------------------------------------------------------------------- #
#                                    Revsymg                                   #
# ---------------------------------------------------------------------------- #
def test_getter_revsymg():
    """Test getter revsymg."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_ALONE_SOL_REGMAP,
        _IR_ALONE_SOL_REGCTG_F,
    )
    # TOTEST: use fixtures and composition
    # TOTEST: find a better assertion
    assert isinstance(asm_graph.revsymg(), RevSymGraph)


# ---------------------------------------------------------------------------- #
#                          Oriented_contigs_of_region                          #
# ---------------------------------------------------------------------------- #
def test_oriented_contigs_of_region():
    """Test oriented_contigs_of_region method."""
    # TOTEST: use fixture for assembly graph
    asm_graph = AssemblyGraph(
        _IR_ALONE_SOL_REGMAP,
        _IR_ALONE_SOL_REGCTG_F,
    )
    # TOTEST: use fixtures and composition
    assert list(asm_graph.oriented_contigs_of_region(0)) == [
        ('C0', FORWARD_INT),
        ('C1', REVERSE_INT),
    ]
    assert list(asm_graph.oriented_contigs_of_region(0, FORWARD_INT)) == [
        ('C0', FORWARD_INT),
        ('C1', REVERSE_INT),
    ]
    assert list(asm_graph.oriented_contigs_of_region(0, REVERSE_INT)) == [
        ('C1', FORWARD_INT),
        ('C0', REVERSE_INT),
    ]


# ---------------------------------------------------------------------------- #
#                              Rev_oriented_contig                             #
# ---------------------------------------------------------------------------- #
def test_rev_oriented_contig():
    """Test rev_oriented_contig function."""
    assert rev_oriented_contig(('0', FORWARD_INT)) == ('0', REVERSE_INT)
    assert rev_oriented_contig(('3', REVERSE_INT)) == ('3', FORWARD_INT)


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def read_region_paths(region_paths_file: Path) -> Iterator[list[IndOrT]]:
    """Read region paths from file.

    Parameters
    ----------
    region_paths_file : Path
        Region paths file path

    Yields
    ------
    list of IndOrT
        Region path
    """
    # REFACTOR: API function?
    with open(region_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            region_path = []
            l_regindor = line.split()
            k = 0
            while k < len(l_regindor) - 1:
                region_path.append(
                    (
                        IndexT(l_regindor[k]),
                        STR_ORIENT[l_regindor[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield region_path


def read_oriented_contig_paths(oriented_contig_paths_file: Path) -> (
        Iterator[list[OrCT]]):
    """Read region paths from file.

    Parameters
    ----------
    oriented_contig_paths_file : Path
        Region paths file path

    Yields
    ------
    list of OrCT
        Oriented contig path
    """
    # REFACTOR: API function?
    with open(oriented_contig_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            oriented_contig_path = []
            l_orc = line.split()
            k = 0
            while k < len(l_orc) - 1:
                oriented_contig_path.append(
                    (
                        l_orc[k],
                        STR_ORIENT[l_orc[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield oriented_contig_path
